from dataclasses import dataclass
from serde import deserialize, serialize
from serde.json import from_json, to_json

# Mark the class serializable/deserializable.
@deserialize
@serialize
@dataclass
class Hoge:
    i: int
    s: str
    f: float
    b: bool


if __name__ == '__main__':
    h = Hoge(i=10, s='hoge', f=100.0, b=True)
    print(f"Into Json: {to_json(h)}")

    s = '{"i": 10, "s": "hoge", "f": 100.0, "b": true}'
    print(f"From Json: {from_json(Hoge, s)}")



import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from serde import deserialize
from serde import init as serde_init
from serde import logger, serialize
from serde.json import from_json, to_json

logging.basicConfig(level=logging.WARNING)
logger.setLevel(logging.DEBUG)

serde_init(True)


@deserialize
@serialize
@dataclass(unsafe_hash=True)
class PriUnion:
    """
    Union Primitives.
    """

    v: Union[int, str, float, bool]


@deserialize
@serialize
@dataclass(unsafe_hash=True)
class PriOptUnion:
    """
    Union Primitives.
    """

    v: Union[Optional[int], Optional[str], Optional[float], Optional[bool]]


@deserialize
@serialize
@dataclass(unsafe_hash=True)
class ContUnion:
    """
    Union Containers.
    """

    v: Union[List[int], List[str], Dict[str, int]]


def test_union():
    v = PriUnion(10)
    s = '{"v": 10}'
    assert s == to_json(v)
    print(f'hoge {v.__serde_hidden__.code}')
    assert v == from_json(PriUnion, s, strict=False)

    v = PriUnion(10.0)
    s = '{"v": 10.0}'
    assert s == to_json(v)
    assert v == from_json(PriUnion, s)

    v = PriUnion('hoge')
    s = '{"v": "hoge"}'
    assert s == to_json(v)
    assert v == from_json(PriUnion, s)

    v = PriUnion(True)
    s = '{"v": true}'
    assert s == to_json(v)
    assert v == from_json(PriUnion, s)


def test_union_optional():
    v = PriOptUnion(10)
    s = '{"v": 10}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion(None)
    s = '{"v": null}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion("hoge")
    s = '{"v": "hoge"}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion(10.0)
    s = '{"v": 10.0}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)

    v = PriOptUnion(False)
    s = '{"v": false}'
    assert s == to_json(v)
    assert v == from_json(PriOptUnion, s)


def test_union_containers():
    v = ContUnion([1, 2, 3])
    s = '{"v": [1, 2, 3]}'
    assert s == to_json(v)
    assert v == from_json(ContUnion, s)

    v = ContUnion(['1', '2', '3'])
    s = '{"v": ["1", "2", "3"]}'
    assert s == to_json(v)
    assert v == from_json(ContUnion, s)

    v = ContUnion({'a': 1, 'b': 2, 'c': 3})
    s = '{"v": {"a": 1, "b": 2, "c": 3}}'
    assert s == to_json(v)
    assert v == from_json(ContUnion, s)



"""
Defines classess and functions for `serialize` decorator.

"""
import abc
import copy  # noqa
import functools
from dataclasses import Field as DataclassField
from dataclasses import asdict as _asdict
from dataclasses import astuple as _astuple
from dataclasses import dataclass
from dataclasses import fields as dataclass_fields  # noqa
from dataclasses import is_dataclass
from typing import Any, Dict, List, Optional, Tuple, Type

import jinja2
import stringcase

from .compat import is_dict, is_list, is_tuple, type_args
from .core import HIDDEN_NAME, SE_NAME, SETTINGS, TO_DICT, TO_ITER, Field, Hidden, T, fields, gen


class Serializer(metaclass=abc.ABCMeta):
    """
    `Serializer` base class. Subclass this to custonize serialize bahaviour.

    See `serde.json.JsonSerializer` and `serde.msgpack.MsgPackSerializer` for example usage.
    """

    @abc.abstractclassmethod
    def serialize(cls, obj):
        pass


def serialize(_cls=None, rename_all: Optional[str] = None) -> Type:
    """
    `serialize` decorator. A dataclass with this decorator can be serialized
    into an object in various data format such as JSON and MsgPack.

    >>> from serde import serialize
    >>> from serde.json import to_json
    >>>
    >>> # Mark the class serializable.
    >>> @serialize
    ... @dataclass
    ... class Hoge:
    ...     i: int
    ...     s: str
    ...     f: float
    ...     b: bool
    >>>
    >>> to_json(Hoge(i=10, s='hoge', f=100.0, b=True))
    '{"i": 10, "s": "hoge", "f": 100.0, "b": true}'
    >>>

    Additionally, `serialize` supports case conversion. Pass case name in
    `serialize` decorator as shown below.

    >>> @serialize(rename_all = 'camelcase')
    ... @dataclass
    ... class Hoge:
    ...     int_field: int
    ...     str_field: str
    >>>
    >>> to_json(Hoge(int_field=10, str_field='hoge'))
    '{"intField": 10, "strField": "hoge"}'
    >>>
    """

    @functools.wraps(_cls)
    def wrap(cls: Type[T]) -> Type[T]:
        if not hasattr(cls, HIDDEN_NAME):
            setattr(cls, HIDDEN_NAME, Hidden())

        def serialize(self, ser, **opts) -> None:
            return ser.serialize(self, **opts)

        setattr(cls, SE_NAME, serialize)

        g: Dict[str, Any] = globals().copy()
        for f in sefields(cls):
            if f.skip_if:
                g[f.skip_if.mangled] = f.skip_if
        cls = se_func(cls, TO_ITER, render_astuple(cls), g)
        cls = se_func(cls, TO_DICT, render_asdict(cls, rename_all), g)
        return cls

    if _cls is None:
        return wrap

    return wrap(_cls)


def is_serializable(instance_or_class: Any) -> bool:
    """
    Test if arg can `serialize`. Arg must be either an instance of class.

    >>> from dataclasses import dataclass
    >>> from serde import serialize, is_serializable
    >>>
    >>> @serialize
    ... @dataclass
    ... class Hoge:
    ...     pass
    >>>
    >>> is_serializable(Hoge)
    True
    >>>
    """
    return hasattr(instance_or_class, '__serde_serialize__')


def to_dict(o) -> Dict:
    """
    Convert object into dictionary.

    >>> @serialize
    ... @dataclass
    ... class Hoge:
    ...     i: int
    >>>
    >>> to_dict(Hoge(10))
    {'i': 10}
    >>>
    >>> to_dict([Hoge(10), Hoge(20)])
    [{'i': 10}, {'i': 20}]
    >>>
    >>> to_dict({'a': Hoge(10), 'b': Hoge(20)})
    {'a': {'i': 10}, 'b': {'i': 20}}
    >>>
    >>> to_dict((Hoge(10), Hoge(20)))
    ({'i': 10}, {'i': 20})
    """
    if o is None:
        v = None
    if is_serializable(o):
        v = asdict(o)
    elif isinstance(o, list):
        v = [to_dict(e) for e in o]
    elif isinstance(o, tuple):
        v = tuple(to_dict(e) for e in o)
    elif isinstance(o, dict):
        v = {k: to_dict(v) for k, v in o.items()}
    else:
        v = o

    return v


def astuple(v):
    """
    Convert `serialize` class to `tuple`.
    """
    if is_serializable(v):
        return getattr(v, TO_ITER)()
    elif is_dataclass(v):
        return _astuple(v)
    elif isinstance(v, Dict):
        return tuple(astuple(e) for e in v.values())
    elif isinstance(v, (Tuple, List)):
        return tuple(astuple(e) for e in v)
    else:
        return v


def asdict(v):
    """
    Convert `serialize` class to `dict`.
    """
    if is_serializable(v):
        return getattr(v, TO_DICT)()
    elif is_dataclass(v):
        return _asdict(v)
    elif isinstance(v, Dict):
        return {asdict(k): asdict(v) for k, v in v.items()}
    elif isinstance(v, (Tuple, List)):
        return tuple(asdict(e) for e in v)
    else:
        return v


@dataclass
class SeField(Field):
    parent: Optional['Field'] = None

    @property
    def varname(self) -> str:
        var = self.parent.varname if self.parent else None
        if var:
            return f'{var}.{self.name}'
        else:
            return self.name

    def __getitem__(self, n) -> 'SeField':
        typ = type_args(self.type)[n]
        return SeField(typ, None)

    @staticmethod
    def mangle(field: DataclassField, name: str) -> str:
        return f'{field.name}_{name}'


sefields = functools.partial(fields, SeField)


def to_arg(f: SeField) -> SeField:
    f.parent = SeField(None, 'obj')
    return f


def render_astuple(cls: Type) -> str:
    template = """
def {{func}}(obj):
  if not is_dataclass(obj):
    return copy.deepcopy(obj)

  {% if cls|is_dataclass %}
  res = []
  {% for f in cls|fields -%}
  {% if not f.skip|default(False) %}
  res.append({{f|arg|rvalue()}})
  {% endif %}
  {% endfor -%}
  return tuple(res)
  {% endif %}
    """

    renderer = Renderer(TO_ITER)
    env = jinja2.Environment(loader=jinja2.DictLoader({'iter': template}))
    env.filters.update({'fields': sefields})
    env.filters.update({'is_dataclass': is_dataclass})
    env.filters.update({'rvalue': renderer.render})
    env.filters.update({'arg': to_arg})
    return env.get_template('iter').render(func=TO_ITER, cls=cls)


def render_asdict(cls: Type, case: Optional[str] = None) -> str:
    template = """
def {{func}}(obj):
  if not is_dataclass(obj):
    return copy.deepcopy(obj)

  {% if cls|is_dataclass %}
  res = {}
  {% for f in cls|fields -%}

  {% if not f.skip %}
    {% if f.skip_if %}
  if not {{f.skip_if.mangled}}({{f|arg|rvalue()}}):
    res["{{f|case}}"] = {{f|arg|rvalue()}}
    {% else %}
  res["{{f|case}}"] = {{f|arg|rvalue()}}
    {% endif %}
  {% endif %}

  {% endfor -%}
  return res
  {% endif %}
    """

    def conv(f: SeField) -> str:
        """
        Convert dict key name.
        """
        name = f.name
        casef = getattr(stringcase, case or '', None)
        if casef:
            name = casef(name)
        if f.rename:
            name = f.rename
        return name

    renderer = Renderer(TO_DICT)
    env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
    env.filters.update({'fields': sefields})
    env.filters.update({'is_dataclass': is_dataclass})
    env.filters.update({'rvalue': renderer.render})
    env.filters.update({'arg': to_arg})
    env.filters.update({'case': conv})
    return env.get_template('dict').render(func=TO_DICT, cls=cls)


@dataclass
class Renderer:
    """
    Render rvalue for variouls types.
    """

    func: str

    def render(self, arg: SeField) -> str:
        """
        Render rvalue

        >>> Renderer(TO_ITER).render(SeField(int, 'i'))
        'i'

        >>> Renderer(TO_ITER).render(SeField(List[int], 'l'))
        '[v for v in l]'

        >>> @serialize
        ... @dataclass(unsafe_hash=True)
        ... class Hoge:
        ...    val: int
        >>> Renderer(TO_ITER).render(SeField(Hoge, 'hoge'))
        'hoge.__serde_to_iter__()'

        >>> Renderer(TO_ITER).render(SeField(List[Hoge], 'hoge'))
        '[v.__serde_to_iter__() for v in hoge]'

        >>> Renderer(TO_ITER).render(SeField(Dict[str, Hoge], 'hoge'))
        '{k: v.__serde_to_iter__() for k, v in hoge.items()}'

        >>> Renderer(TO_ITER).render(SeField(Dict[Hoge, Hoge], 'hoge'))
        '{k.__serde_to_iter__(): v.__serde_to_iter__() for k, v in hoge.items()}'

        >>> Renderer(TO_ITER).render(SeField(Tuple[str, Hoge, int], 'hoge'))
        '(hoge[0], hoge[1].__serde_to_iter__(), hoge[2])'
        >>>
        """
        if is_dataclass(arg.type):
            return self.dataclass(arg)
        elif is_list(arg.type):
            return self.list(arg)
        elif is_dict(arg.type):
            return self.dict(arg)
        elif is_tuple(arg.type):
            return self.tuple(arg)
        else:
            return self.primitive(arg)

    def dataclass(self, arg: SeField) -> str:
        return f'{arg.varname}.{self.func}()'

    def list(self, arg: SeField) -> str:
        earg = arg[0]
        earg.name = 'v'
        return f'[{self.render(earg)} for v in {arg.varname}]'

    def tuple(self, arg: SeField) -> str:
        rvalues = []
        for i, _ in enumerate(type_args(arg.type)):
            r = arg[i]
            r.name = f'{arg.varname}[{i}]'
            rvalues.append(self.render(r))
        return f"({', '.join(rvalues)})"

    def dict(self, arg: SeField) -> str:
        karg = arg[0]
        karg.name = 'k'
        varg = arg[1]
        varg.name = 'v'
        return f'{{{self.render(karg)}: {self.render(varg)} for k, v in {arg.varname}.items()}}'

    def primitive(self, arg: SeField) -> str:
        return f'{arg.varname}'


def se_func(cls: Type[T], func: str, code: str, g: Dict = None, local: Dict = None) -> Type[T]:
    """
    Generate function to serialize into an object.
    """
    # Generate serialize function.
    if not g:
        g = globals().copy()
    if not local:
        local = locals().copy()
    code = gen(code, g, local, cls=cls)

    setattr(cls, func, local[func])
    if SETTINGS['debug']:
        hidden = getattr(cls, HIDDEN_NAME)
        hidden.code[func] = code

    return cls



import enum
import json
import logging
from dataclasses import dataclass, field
from typing import Dict, List

import pytest

from serde import asdict, astuple, deserialize, from_dict
from serde import init as serde_init
from serde import logger, serialize
from serde.json import from_json, to_json
from serde.msgpack import from_msgpack, to_msgpack
from serde.toml import from_toml, to_toml
from serde.yaml import from_yaml, to_yaml

from .data import Bool, Float, Int, NestedPri, NestedPriOpt, NestedPriTuple, Pri, PriOpt, PriTuple, Str

logging.basicConfig(level=logging.WARNING)

logger.setLevel(logging.DEBUG)

serde_init(True)


all_formats = [
    (asdict, from_dict),
    (to_json, from_json),
    (to_msgpack, from_msgpack),
    (to_yaml, from_yaml),
    (to_toml, from_toml),
]


@pytest.mark.parametrize('se,de', all_formats)
def test_primitive(se, de):
    p = Pri(10, 'foo', 100.0, True)
    assert p == de(Pri, se(p))


@pytest.mark.parametrize('se,de', all_formats)
def test_nested_primitive(se, de):
    p = NestedPri(Int(10), Str('foo'), Float(100.0), Bool(True))
    assert p == from_dict(NestedPri, asdict(p))


def test_non_dataclass():
    with pytest.raises(TypeError):

        @deserialize
        @serialize
        class Hoge:
            i: int


def test_forward_declaration():
    @serialize
    @deserialize
    @dataclass
    class Foo:
        bar: 'Bar'

    @serialize
    @deserialize
    @dataclass
    class Bar:
        i: int

    h = Foo(bar=Bar(i=10))
    assert h.bar.i == 10


def test_enum():
    class IEnum(enum.IntEnum):
        V0 = enum.auto()
        V1 = enum.auto()
        V2 = enum.auto()

    class SEnum(enum.Enum):
        V0 = 'v0'
        V1 = 'v1'
        V2 = 'v2'

    # NOTE: Only IntEnum is supported now.
    @deserialize
    @serialize
    @dataclass
    class Hoge:
        ie0: IEnum
        ie1: IEnum
        ie2: IEnum
        se0: SEnum
        se1: SEnum
        se2: SEnum


@pytest.mark.parametrize('se,de', all_formats)
def test_tuple(se, de):
    p = PriTuple(
        (10, 20, 30), ('a', 'b', 'c', 'd'), (10.0, 20.0, 30.0, 40.0, 50.0), (True, False, True, False, True, False)
    )
    tpl: PriTuple = de(PriTuple, se(p))
    assert tpl.i == (10, 20, 30)
    assert tpl.s == ('a', 'b', 'c', 'd')
    assert tpl.f == (10.0, 20.0, 30.0, 40.0, 50.0)
    assert tpl.b == (True, False, True, False, True, False)

    # List can also be used.
    p = PriTuple(
        [10, 20, 30], ['a', 'b', 'c', 'd'], [10.0, 20.0, 30.0, 40.0, 50.0], [True, False, True, False, True, False]
    )
    tpl: PriTuple = de(PriTuple, se(p))
    assert tpl.i == (10, 20, 30)
    assert tpl.s == ('a', 'b', 'c', 'd')
    assert tpl.f == (10.0, 20.0, 30.0, 40.0, 50.0)
    assert tpl.b == (True, False, True, False, True, False)


@pytest.mark.parametrize('se,de', all_formats)
def test_dataclass_in_tuple(se, de):
    src = NestedPriTuple(
        (Int(10), Int(10), Int(10)),
        (Str("10"), Str("10"), Str("10"), Str("10")),
        (Float(10.0), Float(10.0), Float(10.0), Float(10.0), Float(10.0)),
        (Bool(False), Bool(False), Bool(False), Bool(False), Bool(False), Bool(False)),
    )
    assert src == from_json(NestedPriTuple, to_json(src))

    with pytest.raises(IndexError):
        j = json.dumps(
            {
                'i': (10, 20),
                's': ('a', 'b', 'c', 'd'),
                'f': (10.0, 20.0, 30.0, 40.0, 50.0),
                'b': (True, False, True, False, True, False),
            }
        )
        _: PriTuple = from_json(PriTuple, j)


@pytest.mark.parametrize('se,de', all_formats)
def test_optional(se, de):
    p = PriOpt(20, None, 100.0, True)
    assert p == de(PriOpt, se(p))


def test_optional_dataclass():
    pass
    # p = NestedPriOpt(Int(20), Str(None), Float(100.0), Bool(True))
    # s = '{"i": 20, "s": null, "f": 100.0, "b": true}'
    # TODO
    # assert s == asdict(p)
    # assert p == from_json(PriOpt, s)


def test_complex():
    @deserialize
    @serialize
    @dataclass
    class Foo:
        v: List[int] = field(default_factory=list)
        d: Dict[str, int] = field(default_factory=dict)

    @deserialize
    @serialize
    @dataclass
    class Bar:
        i: int

    @deserialize
    @serialize
    @dataclass
    class Hoge:
        i: int
        s: str
        f: float
        b: bool
        foo: Foo
        lst: List[Bar] = field(default_factory=list)
        lst2: List[Dict[str, Bar]] = field(default_factory=list)
        dct: Dict[str, List[List[Bar]]] = field(default_factory=dict)

    f = Foo(v=[1, 2, 3, 4, 5], d={'hoge': 10, 'fuga': 20})
    lst = [Bar(10), Bar(20)]
    lst2 = [{'bar1': Bar(10)}, {'bar2': Bar(10), 'bar3': Bar(20)}]
    dct = {'hoge': [[Bar(10), Bar(20)], [Bar(20), Bar(30)]]}
    h = Hoge(i=10, s='hoge', f=100.0, b=True, foo=f, lst=lst, lst2=lst2, dct=dct)
    s = """
               {"i": 10,
                "s": "hoge",
                "f": 100.0,
                "b": true,
                "foo" : {
                    "v": [1, 2, 3, 4, 5],
                    "d": {"hoge": 10, "fuga": 20}
                },
                "lst": [{"i": 10}, {"i": 20}],
                "lst2": [{"bar1": {"i": 10}}, {"bar2": {"i": 10}, "bar3": {"i": 20}}],
                "dct": {"hoge": [[{"i": 10}, {"i": 20}], [{"i": 20}, {"i": 30}]]}
                }
                """
    assert json.loads(s) == json.loads(to_json(h))
    hh = from_json(Hoge, s)
    assert h.foo == hh.foo
    assert h.lst == hh.lst
    assert h.dct == hh.dct


def test_json():
    p = Pri(10, 'foo', 100.0, True)
    s = '{"i": 10, "s": "foo", "f": 100.0, "b": true}'
    assert s == to_json(p)

    assert '10' == to_json(10)
    assert '[10, 20, 30]' == to_json([10, 20, 30])
    assert '{"hoge": 10, "fuga": 10}' == to_json({'hoge': 10, 'fuga': 10})


def test_msgpack():
    p = Pri(10, 'foo', 100.0, True)
    d = b'\x84\xa1i\n\xa1s\xa3foo\xa1f\xcb@Y\x00\x00\x00\x00\x00\x00\xa1b\xc3'
    assert d == to_msgpack(p)
    assert p == from_msgpack(Pri, d)



"""
Defines classess and functions for `deserialize` decorator.

`deserialize` is a decorator to make a `dataclasses.dataclass` class deserializable.
`is_deserializable` is used to test a class is with `deserialize`.
`Deserializer` is a deserializer base class used in `from_obj`,
`serde.json.from_json`. You can subclass it to make your own deserializer.
`from_obj` deserializes from an object into an instance of the class with
`deserialize`.

`args_from_iter` and `args_from_dict` are private functions but they are the core
parts of pyserde.
"""
import abc
import functools
from dataclasses import dataclass, is_dataclass
from typing import Any, Dict, List, Optional, Tuple, Type

import jinja2
import stringcase

from .compat import assert_type, is_dict, is_list, is_opt, is_tuple, is_union, iter_types, type_args
from .core import FROM_DICT, FROM_ITER, HIDDEN_NAME, SETTINGS, Field, Hidden, SerdeError, T, fields, gen


def deserialize(_cls=None, rename_all: Optional[str] = None) -> Type:
    """
    `deserialize` decorator. A dataclass with this decorator can be deserialized
    into an object from various data format such as JSON and MsgPack.

    >>> from serde import deserialize
    >>> from serde.json import from_json
    >>>
    >>> # Mark the class deserializable.
    >>> @deserialize
    ... @dataclass
    ... class Hoge:
    ...     i: int
    ...     s: str
    ...     f: float
    ...     b: bool
    >>>
    >>> from_json(Hoge, '{"i": 10, "s": "hoge", "f": 100.0, "b": true}')
    Hoge(i=10, s='hoge', f=100.0, b=True)

    Additionally, `deserialize` supports case conversion. Pass case name in
    `deserialize` decorator as shown below.

    >>> from serde import deserialize
    >>>
    >>> @deserialize(rename_all = 'camelcase')
    ... @dataclass
    ... class RenameAll:
    ...     int_field: int
    ...     str_field: str
    >>>
    >>> from_json(RenameAll, '{"intField": 10, "strField": "hoge"}')
    RenameAll(int_field=10, str_field='hoge')
    >>>
    """

    def wrap(cls):
        if not hasattr(cls, HIDDEN_NAME):
            setattr(cls, HIDDEN_NAME, Hidden())
        cls = de_func(cls, FROM_ITER, render_from_iter(cls))
        cls = de_func(cls, FROM_DICT, render_from_dict(cls, rename_all))
        return cls

    if _cls is None:
        return wrap

    return wrap(_cls)


def is_deserializable(instance_or_class: Any) -> bool:
    """
    Test if arg can `deserialize`. Arg must be either an instance of class.

    >>> from serde import deserialize, is_deserializable
    >>>
    >>> @deserialize
    ... @dataclass
    ... class Hoge:
    ...     pass
    >>>
    >>> is_deserializable(Hoge)
    True
    >>>
    """
    return hasattr(instance_or_class, FROM_ITER) or hasattr(instance_or_class, FROM_DICT)


class Deserializer(metaclass=abc.ABCMeta):
    """
    `Deserializer` base class. Subclass this to custonize deserialize bahaviour.

    See `serde.json.JsonDeserializer` and `serde.msgpack.MsgPackDeserializer` for example usage.
    """

    @abc.abstractclassmethod
    def deserialize(cls, data, **opts):
        """
        deserialize `data` into an object typically `dict`, `list` or `tuple`.

        For example, `serde.json.JsonDeserializer` takes json string and deserialize
        into an object. `serde.msgpack.MsgPackDeserializer` takes msgpack bytes and
        deserialize into an object.
        """


def from_obj(c: Type[T], o: Any, de: Type[Deserializer] = None, strict=True, **opts):
    """
    Deserialize from an object into an instance of the type specified as arg `c`.
    `c` can be either primitive type, `List`, `Tuple`, `Dict` or `deserialize` class.

    ### Dataclass

    # >>> from serde import deserialize
    # >>>
    # >>> @deserialize
    # ... @dataclass
    # ... class Hoge:
    # ...     i: int
    # ...     f: float
    # ...     s: str
    # ...     b: bool
    # >>>
    # >>> obj = {'i': 10, 'f': 0.1, 's': 'hoge', 'b': False}
    # >>> from_obj(Hoge, obj)
    # Hoge(i=10, f=0.1, s='hoge', b=False)

    # ### Containers

    # >>> from serde import deserialize
    # >>>
    # >>> @deserialize
    # ... @dataclass
    # ... class Hoge:
    # ...     s: str
    # ...     i: int
    # >>>
    # >>> from_obj(List[Hoge], [('hoge', 10), ('foo', 20)])
    # [Hoge(s='hoge', i=10), Hoge(s='foo', i=20)]
    # >>>
    # >>> from_obj(Dict[str ,Hoge], {'hoge': ('hoge', 10), 'foo': ('foo', 20)})
    # {'hoge': Hoge(s='hoge', i=10), 'foo': Hoge(s='foo', i=20)}
    # >>>
    """
    if de:
        o = de.deserialize(o, **opts)
    if o is None:
        v = None
    if is_deserializable(c):
        v = from_dict_or_iter(c, o)
    elif is_opt(c):
        if o is None:
            v = None
        else:
            v = from_obj(type_args(c)[0], o)
    elif is_union(c):
        v = None
        for typ in type_args(c):
            try:
                v = from_obj(typ, o)
                break
            except (SerdeError, ValueError):
                pass
    elif is_list(c):
        assert_type(list, o, strict)
        v = [from_obj(type_args(c)[0], e) for e in o]
    elif is_tuple(c):
        assert_type(tuple, o, strict)
        v = tuple(from_obj(type_args(c)[i], e) for i, e in enumerate(o))
    elif is_dict(c):
        assert_type(dict, o, strict)
        v = {from_obj(type_args(c)[0], k): from_obj(type_args(c)[1], v) for k, v in o.items()}
    else:
        v = o

    return v


def from_dict_or_iter(cls, o):
    """
    Deserialize into an instance of `deserialize` class from either `dict` or `iterable`.
    """
    if not is_deserializable(cls):
        raise SerdeError('`cls` must be deserializable.')

    if isinstance(o, (List, Tuple)):
        return cls.__serde_from_iter__(o)
    elif isinstance(o, Dict):
        return cls.__serde_from_dict__(o)
    elif isinstance(o, cls) or o is None:
        return o
    else:
        raise SerdeError(f'Arg must be either List, Tuple, Dict or Type but {type(o)}.')


def from_dict(cls, o):
    return cls.__serde_from_dict__(o)


def from_tuple(cls, o):
    return cls.__serde_from_tuple__(o)


def case(s: str, case: Optional[str]) -> str:
    if case:
        f = getattr(stringcase, case, None)
        if not f:
            raise SerdeError((f"Unkown case type: {case}. Pass the name of case supported by 'stringcase' package."))
        return f(s)
    else:
        return s


@dataclass
class DeField(Field):
    datavar: Optional[str] = None  # name of variable to deserialize from.

    def __getitem__(self, n) -> 'DeField':
        typ = type_args(self.type)[n]
        if is_list(self.type) or is_dict(self.type):
            return ElementField(typ, 'v', datavar='v')
        elif is_tuple(self.type):
            return ElementField(typ, f'{self.data}[{n}]', datavar=f'{self.data}[{n}]')
        else:
            return DeField(typ, self.name, datavar=self.datavar)

    def get_kv(self) -> Tuple['ElementField', 'ElementField']:
        k = self[0]
        k.name = 'k'
        k.datavar = 'k'
        v = self[1]
        return (k, v)

    @property
    def data(self) -> str:
        return f'{self.datavar}["{case(self.name, self.case)}"]'

    @data.setter
    def data(self, d):
        self.datavar = d


@dataclass
class ElementField(DeField):
    """
    Field for element type such as List[T].
    """

    @property
    def data(self) -> str:
        return self.datavar

    @data.setter
    def data(self, d):
        self.datavar = d


defields = functools.partial(fields, DeField)


@dataclass
class Renderer:
    """
    Base Renderer class.
    """

    func: str

    def render(self, arg: DeField) -> str:
        """
        Render rvalue
        """
        if is_dataclass(arg.type):
            return self.dataclass(arg)
        elif is_opt(arg.type):
            return self.opt(arg)
        elif is_list(arg.type):
            return self.list(arg)
        elif is_dict(arg.type):
            return self.dict(arg)
        elif is_tuple(arg.type):
            return self.tuple(arg)
        else:
            return self.primitive(arg)

    def dataclass(self, arg: DeField) -> str:
        """ Render rvalue for dataclass. """

    def opt(self, arg: DeField) -> str:
        """ Render rvalue for optional. """

    def list(self, arg: DeField) -> str:
        """ Render rvalue for list. """

    def tuple(self, arg: DeField) -> str:
        """ Render rvalue for tuple. """

    def dict(self, arg: DeField) -> str:
        """ Render rvalue for dict. """

    def primitive(self, arg: DeField) -> str:
        """ Render rvalue for primitives. """


@dataclass
class DictRenderer(Renderer):
    """
    Render rvalue for various types.
    """

    def dataclass(self, arg: DeField) -> str:
        return f'{arg.type.__name__}.{self.func}({arg.data})'

    def opt(self, arg: DeField) -> str:
        """
        Render rvalue for Optional.

        >>> DictRenderer('hoge').render(DeField(Optional[int], 'o', datavar='data'))
        'data["o"] if "o" in data else None'

        >>> DictRenderer('hoge').render(DeField(Optional[List[int]], 'o', datavar='data'))
        '[v for v in data["o"]] if "o" in data else None'

        >>> DictRenderer('hoge').render(DeField(Optional[List[int]], 'o', datavar='data'))
        '[v for v in data["o"]] if "o" in data else None'

        >>> @deserialize
        ... @dataclass
        ... class Foo:
        ...     o: Optional[List[int]]
        >>> DictRenderer('hoge').render(DeField(Optional[Foo], 'f', datavar='data'))
        'Foo.hoge(data["f"]) if "f" in data else None'
        """
        value = arg[0]
        exists = f'"{value.name}" in {value.datavar}'
        return f'{self.render(value)} if {exists} else None'

    def list(self, arg: DeField) -> str:
        """
        Render rvalue for list.

        >>> DictRenderer('hoge').render(DeField(List[int], 'l', datavar='data'))
        '[v for v in data["l"]]'

        >>> DictRenderer('hoge').render(DeField(List[List[int]], 'l', datavar='data'))
        '[[v for v in v] for v in data["l"]]'
        """
        return f'[{self.render(arg[0])} for v in {arg.data}]'

    def tuple(self, arg: DeField) -> str:
        """
        Render rvalue for tuple.

        >>> @deserialize
        ... @dataclass
        ... class Foo: pass
        >>> DictRenderer('hoge').render(DeField(Tuple[str, int, List[int], Foo], 'd', datavar='data'))
        '(data["d"][0], data["d"][1], [v for v in data["d"][2]], Foo.hoge(data["d"][3]))'
        """
        values = []
        for i, typ in enumerate(type_args(arg.type)):
            inner = arg[i]
            values.append(self.render(inner))
        return f'({", ".join(values)})'

    def dict(self, arg: DeField) -> str:
        """
        Render rvalue for dict.

        >>> DictRenderer('hoge').render(DeField(Dict[str, int], 'd', datavar='data'))
        '{k: v for k, v in data["d"].items()}'

        >>> @deserialize
        ... @dataclass
        ... class Foo: pass
        >>> DictRenderer('hoge').render(DeField(Dict[Foo, List[Foo]], 'f', datavar='data'))
        '{Foo.hoge(k): [Foo.hoge(v) for v in v] for k, v in data["f"].items()}'
        """
        k, v = arg.get_kv()
        return f'{{{self.render(k)}: {self.render(v)} for k, v in {arg.data}.items()}}'

    def primitive(self, arg: DeField) -> str:
        """
        Render rvalue for primitives.

        >>> DictRenderer('hoge').render(DeField(int, 'i', datavar='data'))
        'data["i"]'

        >>> DictRenderer('hoge').render(DeField(int, 'int_field', datavar='data', case='camelcase'))
        'data["intField"]'
        """
        return arg.data


def to_arg(f: Field, rename_all: Optional[str] = None) -> DeField:
    f.data = 'data'
    f.parent = DeField(None, 'data', datavar='data', case=f.case or rename_all)
    f.case = f.case or rename_all
    return f


def render_from_iter(cls: Type) -> str:
    template = """
def {{func}}(data):
  return cls(
  {% for f in cls|fields %}
  {{f|arg|rvalue}},
  {% endfor %}
  )
    """

    renderer = DictRenderer(FROM_ITER)
    env = jinja2.Environment(loader=jinja2.DictLoader({'iter': template}))
    env.filters.update({'rvalue': renderer.render})
    env.filters.update({'fields': defields})
    env.filters.update({'arg': to_arg})
    return env.get_template('iter').render(func=FROM_ITER, cls=cls)


def render_from_dict(cls: Type, rename_all: Optional[str] = None) -> str:
    template = """
def {{func}}(data):
  return cls(
  {% for f in cls|fields %}
  {{f|arg|rvalue}},
  {% endfor %}
  )
    """

    renderer = DictRenderer(FROM_DICT)
    env = jinja2.Environment(loader=jinja2.DictLoader({'dict': template}))
    env.filters.update({'rvalue': renderer.render})
    env.filters.update({'fields': defields})
    env.filters.update({'arg': functools.partial(to_arg, rename_all=rename_all)})
    return env.get_template('dict').render(func=FROM_DICT, cls=cls)


def de_func(cls: Type[T], func: str, code: str) -> Type[T]:
    """
    Generate function to deserialize into an instance of `deserialize` class.
    """
    # Collect types to be used in the `exec` scope.
    g: Dict[str, Any] = globals().copy()
    for typ in iter_types(cls):
        if is_dataclass(typ):
            g[typ.__name__] = typ
    g['cls'] = cls
    import typing

    g['typing'] = typing
    g['NoneType'] = type(None)

    # Generate deserialize function.
    code = gen(code, g, cls=cls)
    setattr(cls, func, staticmethod(g[func]))
    if SETTINGS['debug']:
        hidden = getattr(cls, HIDDEN_NAME)
        hidden.code[func] = code

    return cls
